package org.mcmas.ui.editors;


import java.text.MessageFormat;
import java.util.ArrayList;

import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.ITextViewer;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.TextPresentation;
import org.eclipse.jface.text.contentassist.*;
import org.eclipse.jface.text.rules.IToken;
import org.mcmas.ui.syntax.Util;

/**
 * ISPL completion processor.
 */
public class IsplCompletionProcessor implements IContentAssistProcessor {
	MCMASEditor editor;
	
	public IsplCompletionProcessor(MCMASEditor editor) {
		super();
		this.editor = editor;
	}
	/**
	 * Simple content assist tip closer. The tip is valid in a range
	 * of 5 characters around its popup location.
	 */
	protected static class Validator implements IContextInformationValidator, IContextInformationPresenter {

		protected int fInstallOffset;

		/*
		 * @see IContextInformationValidator#isContextInformationValid(int)
		 */
		public boolean isContextInformationValid(int offset) {
			return Math.abs(fInstallOffset - offset) < 5;
		}

		/*
		 * @see IContextInformationValidator#install(IContextInformation, ITextViewer, int)
		 */
		public void install(IContextInformation info, ITextViewer viewer, int offset) {
			fInstallOffset= offset;
		}
		
		/*
		 * @see org.eclipse.jface.text.contentassist.IContextInformationPresenter#updatePresentation(int, TextPresentation)
		 */
		public boolean updatePresentation(int documentPosition, TextPresentation presentation) {
			return false;
		}
	}

	protected final static String[] fgProposals=
		{ "Action", "Actions", "Agent", "end", "Environment", "Evaluation", 
		  "Evolution", "Fairness", "Formulae", "Groups", "InitStates", "Obsvars", 
		  "Protocol", "RedStates", "Vars", "Other", "if", "and", "or", "AG", "EG", 
		  "AX", "EX", "X", "F", "G", "AF", "EF", "A", "E" , "U", "K", "GK", "GCK", 
		  "O", "KH", "DKH", "DK", "boolean", "false", "true" };
	protected IContextInformationValidator fValidator= new Validator();

	
	private CompletionProposal generateProposal(String str, String displaystr, int documentOffset) {
		IContextInformation info= new ContextInformation(str, 
				MessageFormat.format(MCMASEditorMessages.getString("CompletionProcessor.Proposal.ContextInfo.pattern"), new Object[] { str })); //$NON-NLS-1$
		return new CompletionProposal(str, documentOffset, 0, str.length(), null, displaystr, info, 
				MessageFormat.format(MCMASEditorMessages.getString("CompletionProcessor.Proposal.hoverinfo.pattern"), new Object[] { str}));
	}
	
	/*private boolean isIdentity(String str) {
		if(str != null && !str.isEmpty()) {
			if(!Character.isLetter(str.charAt(0)))
				return false;
			for(int i=1; i<str.length(); i++)
				if(!(Character.isLetterOrDigit(str.charAt(i)) || c == '_'))
					return false;
			return true;
		}
		return false;
	}*/
	
	private String[] getActions(IDocument document, Segment agent) {
		String[] result = null;
		ArrayList<String> actionList = new ArrayList<String>();
		if(agent.children != null)
			for(int i=0; i<agent.children.size(); i++) {
				Segment actions = agent.children.get(i);
				if(actions.name.compareTo("Actions") == 0) {
					try{
						String str = document.get(actions.position.offset, actions.position.length);
						int openBracket = str.indexOf('{');
						if(openBracket < 0)
							openBracket = str.indexOf("=");
						if(openBracket > 0) {
							IsplWordScanner scanner = new IsplWordScanner();
							scanner.setRange(document, actions.position.offset+openBracket+1, str.length()-openBracket-1);
							IToken t = scanner.nextToken();
							while(!t.isEOF()) {
								if(t.isOther() && t.getData() != null && t.getData().toString().compareTo(IsplWordScanner.ISPL_WORD) == 0) 
									actionList.add(document.get(scanner.getTokenOffset(), scanner.getTokenLength()));
								t = scanner.nextToken();
							}
							if(actionList.size() > 0) {
								result = new String[actionList.size()];
								actionList.toArray(result);
							}
						}
					}catch(BadLocationException e) {
						
					}
				}
			}
		return result;
	}
	
	/* (non-Ispldoc)
	 * Method declared on IContentAssistProcessor
	 */
	public ICompletionProposal[] computeCompletionProposals(ITextViewer viewer, int documentOffset) {
		ICompletionProposal[] result = null;
		ArrayList<ICompletionProposal> proposals = new ArrayList<ICompletionProposal>();
		IDocument document = viewer.getDocument();
		try{
			Segment[] path = editor.getTreePath(documentOffset);
			if(path==null)
				return null;
			boolean isInAgent = path[0].name.startsWith("Agent");
			int dotOffset = Util.getDotOffset(document, documentOffset);
			if(dotOffset > -1) {
				Position  p = Util.getWordPositionBackward(document, documentOffset);
				String str = document.get(p.offset, p.length);
				Segment seg = editor.getAgentSegmentbyName(str);
				if(seg != null) {
					if(isInAgent) { // search obsvars in the environment
						proposals.add(generateProposal("Action", "Action", documentOffset));
						if(str.compareTo("Environment") == 0) {
							if(seg.children!=null)
								for(int i=0; i<seg.children.size(); i++) {
									Segment obsvars = seg.children.get(i);
									if(obsvars.name.compareTo("Obsvars") == 0) {
										if(obsvars.children != null) 
											for(int j=0; j<obsvars.children.size(); j++) {
												Segment variable = obsvars.children.get(j);
												int spaceOffset = variable.name.indexOf(' ');
												if(spaceOffset > 0) {
													String varName = variable.name.substring(0, spaceOffset);
													proposals.add(generateProposal(varName, variable.name, documentOffset));
												}
											}
										break;
									}
								}
						}
						result = new ICompletionProposal[proposals.size()];
						proposals.toArray(result);
						return result;
					} else { // search all variables
						if (seg.children != null)
							for (int i = 0; i < seg.children.size(); i++) {
								Segment obsvars = seg.children.get(i);
								if ((obsvars.name.compareTo("Obsvars") == 0 && str.compareTo("Environment") == 0)
										|| obsvars.name.compareTo("Vars") == 0) {
									if (obsvars.children != null)
										for (int j = 0; j < obsvars.children.size(); j++) {
											Segment variable = obsvars.children.get(j);
											int spaceOffset = variable.name.indexOf(' ');
											if (spaceOffset > 0) {
												String varName = variable.name.substring(0, spaceOffset);
												proposals.add(generateProposal(varName, variable.name, documentOffset));
											}
										}
								}
							}
						if(proposals.size() > 0) {
							result = new ICompletionProposal[proposals.size()];
							proposals.toArray(result);
							return result;
						}
					}
				}
			} else {
				int eqOffset = Util.getEqOffset(document, documentOffset);
				if(eqOffset > -1) {
					Position  p = Util.getWordPositionBackward(document, eqOffset);
					String str = document.get(p.offset, p.length);
					if(str.compareTo("Action") == 0) {
						dotOffset = Util.getDotOffset(document, p.offset);
						if(dotOffset > -1) { // global action
							p = Util.getWordPositionBackward(document, p.offset);
							str = document.get(p.offset, p.length);
							Segment seg = editor.getAgentSegmentbyName(str);
							if(seg != null) {
								String[] actionList = getActions(document, seg);
								if(actionList != null) {
									for(int j=0; j<actionList.length; j++) 
										proposals.add(generateProposal(actionList[j], actionList[j], documentOffset));
									result = new ICompletionProposal[proposals.size()];
									proposals.toArray(result);
									return result;
								}
							}
						} else { // local action
							if(isInAgent) {
								String[] actionList = getActions(document, path[0]);
								if(actionList != null) {
									for(int j=0; j<actionList.length; j++) 
										proposals.add(generateProposal(actionList[j], actionList[j], documentOffset));
									result = new ICompletionProposal[proposals.size()];
									proposals.toArray(result);
									return result;
								}
							}
						}
					}
				}
			}
		}catch(BadLocationException e) {
			
		}
		
		return null;
		//for (int i= 0; i < fgProposals.length; i++) {
		//	IContextInformation info= new ContextInformation(fgProposals[i], MessageFormat.format(/*IsplEditorMessages.getString*/("CompletionProcessor.Proposal.ContextInfo.pattern"), new Object[] { fgProposals[i] })); //$NON-NLS-1$
		//	result[i]= new CompletionProposal(fgProposals[i], documentOffset, 0, fgProposals[i].length(), null, fgProposals[i], info, MessageFormat.format(/*IsplEditorMessages.getString*/("CompletionProcessor.Proposal.hoverinfo.pattern"), new Object[] { fgProposals[i]})); //$NON-NLS-1$
		//}
		//return result; 
	}
	
	/* (non-Ispldoc)
	 * Method declared on IContentAssistProcessor
	 */
	public IContextInformation[] computeContextInformation(ITextViewer viewer, int documentOffset) {
		IContextInformation[] result= new IContextInformation[5];
		for (int i= 0; i < result.length; i++)
			result[i]= new ContextInformation(
				MessageFormat.format(MCMASEditorMessages.getString("CompletionProcessor.ContextInfo.display.pattern"), new Object[] { new Integer(i), new Integer(documentOffset) }),  //$NON-NLS-1$
				MessageFormat.format(MCMASEditorMessages.getString("CompletionProcessor.ContextInfo.value.pattern"), new Object[] { new Integer(i), new Integer(documentOffset - 5), new Integer(documentOffset + 5)})); //$NON-NLS-1$
		return result;
	}
	
	/* (non-Ispldoc)
	 * Method declared on IContentAssistProcessor
	 */
	public char[] getCompletionProposalAutoActivationCharacters() {
		return new char[] { '.', '=' };
	}
	
	/* (non-Ispldoc)
	 * Method declared on IContentAssistProcessor
	 */
	public char[] getContextInformationAutoActivationCharacters() {
		return new char[] { '#' };
	}
	
	/* (non-Ispldoc)
	 * Method declared on IContentAssistProcessor
	 */
	public IContextInformationValidator getContextInformationValidator() {
		return fValidator;
	}
	
	/* (non-Ispldoc)
	 * Method declared on IContentAssistProcessor
	 */
	public String getErrorMessage() {
		return null;
	}
}
